# OME/ImportExport/ModuleImport.pm

#-------------------------------------------------------------------------------
#
# Copyright (C) 2003 Open Microscopy Environment
#       Massachusetts Institute of Technology,
#       National Institutes of Health,
#       University of Dundee
#
#
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#-------------------------------------------------------------------------------




#-------------------------------------------------------------------------------
#
# Written by:    Josiah Johnston <siah@nih.gov>
#
#-------------------------------------------------------------------------------


package OME::ImportExport::ModuleImport;

use XML::LibXML;
use OME::Tasks::LSIDManager;
use OME::Tasks::ModuleTasks; # for help dealing with Module categories
use OME::Session;
use Log::Agent;
use strict;

=head1 NAME

OME::ImportExport::ModuleImport - Import an Analysis Module XML specification.

=head1 SYNOPSIS

	use OME::ImportExport::ModuleImport;
	use OME::SessionManager;

	my $manager       = OME::SessionManager->new();
	my $session       = $manager->TTYlogin();
	my $programImport = OME::ImportExport::ModuleImport->new( 
		_parser => $parser
	);

	my $newPrograms   = $programImport->importXMLFile( $filePath );

=head1 DESCRIPTION

This module automates the module import process. Given an XML specification
of a module, this will import it into the OME system.
Specifically, it will:
install the module onto the local system
register the module with the database
add any custom tables & columns (to the DB) that the module requires

=cut

# FIXME: Should verify that every table and column declared are used. 


sub new {
	my $proto  = shift;
	my $class  = ref($proto) || $proto;
	my %params = @_;
	
	my @requiredParams = ('session' );
	
	my $self = {
		_parser => $params{_parser},
	};
	
	if (!defined $self->{_parser}) {
		my $parser = XML::LibXML->new();
		die "Cannot create XML parser"
		  unless defined $parser;
		
		$parser->validation(exists $params{ValidateXML}?
							$params{ValidateXML}: 0);
		$self->{_parser} = $parser;
	}

	bless($self,$class);
	return $self;
}


=pod

=head2 importXMLFile

Input Parameters:
	filePath - path to xml file to import

Description:
	Import a module from its description in an xml file

=cut

###############################################################################
#
# parameter(s) are:
#	$filePath
#
sub importXMLFile {
	my $self       = shift;
	my $filePath   = shift;

	my $session    = OME::Session->instance();

#	logdbg "debug", ref ($self) . "->importXMLFile called with parameters:\n\t[filePath=] $filePath"
	my $parser = $self->{_parser};

	#Parse
	my $tree = $parser->parse_file( $filePath )
		or die ref($self) . " Could not parse file ($filePath)";

	#process tree
	my $newPrograms = $self->processDOM( $tree->getDocumentElement() );

	#return a list of imported programs (OME::Modules objects)
	return $newPrograms;
}
#
#
###############################################################################


###############################################################################
#
# Process DOM tree
# parameters:
#	$root element (DOM model)
# returns:
#	list of imported programs
#
sub processDOM {
	my $self    = shift;
	my $root    = shift;

	my $session = OME::Session->instance();
	my $factory = $session->Factory();
	my $lsidManager = OME::Tasks::LSIDManager->new();

	my @commitOnSuccessfulImport;
	my @newPrograms;
	logdbg "debug", ref ($self) . "->processDOM called to process " . scalar(@{$root->getElementsByLocalName( "AnalysisModule" )} ) . " modules";


	foreach my $categoryXML ($root->getElementsByLocalName('Category')) {
		my $categoryPath = $categoryXML->getAttribute('Path');
		my $categoryDescriptions = $categoryXML->getElementsByLocalName('Description');
		my $categoryDescription = [$categoryDescriptions->[0]->childNodes()]->[0]->data()
			if $categoryDescriptions and [$categoryDescriptions->[0]->childNodes()]->[0];

		OME::Tasks::ModuleTasks::makeCategoriesFromPath($categoryPath,$categoryDescription);
	}


foreach my $moduleXML ($root->getElementsByLocalName( "AnalysisModule" )) {

	###########################################################################
	#
	# make OME::Modules object
	#
	logdbg "debug", ref ($self) . "->processDOM creating OME::Module ".$moduleXML->getAttribute( 'ModuleName' );
	my $program = $lsidManager->getObject( $moduleXML->getAttribute( 'ID' ) );
	if (defined $program) {
		print STDERR ref ($self).": Module ID '".$moduleXML->getAttribute( 'ID' ).
			"' exists in DB - ignored\n" if $OME::MESSAGES{LSID_COLLISION};
		logdbg "debug", ref($self)."->processDOM: ". $moduleXML->getAttribute( 'ModuleName' ) ." already exists in the DB. Skipping...";
		next;
	}
	
	my $categoryPath = $moduleXML->getAttribute('Category');
	my $categoryID = OME::Tasks::ModuleTasks::makeCategoriesFromPath($categoryPath)->id()
		if defined $categoryPath;

	my $descriptions = $moduleXML->getElementsByLocalName('Description');
	my $description = [$descriptions->[0]->childNodes()]->[0]->data()
		if $descriptions and $descriptions->[0]->childNodes()->size() > 0;
	my $module_type = $moduleXML->getAttribute( 'ModuleType' );
	my $data = {
		name     => $moduleXML->getAttribute( 'ModuleName' ),
		description      => $description,
		category         => $categoryID,
		module_type      => $module_type,
		# location using ProgramID attribute is a temporary hack
		location         => $moduleXML->getAttribute( 'ProgramID' ),
		default_iterator => $moduleXML->getAttribute( 'FeatureIterator' ),
		new_feature_tag  => $moduleXML->getAttribute( 'NewFeatureName' ),
	};
	my $newProgram = $factory->newObject("OME::Module",$data)
		or die "Could not create OME::Module object\n";
	push(@commitOnSuccessfulImport, $newProgram);
	$lsidManager->setLSID( $newProgram, $moduleXML->getAttribute( 'ID' ) )
		or die "Couldn't set LSID for module ".$moduleXML->getAttribute( 'ID' );
	#
	#
	###########################################################################
	
	
	##########################################################################
	#
	# process formalInputs 
	#
	# this hash is keyed by FormalInput.Name, valued by DBObject FormalInput
	my %formalInputs;

	logdbg "debug", ref ($self) . "->processDOM processing formal inputs";
	foreach my $formalInputXML ( $moduleXML->getElementsByLocalName( "FormalInput" ) ) {
		logdbg "debug", "\tformal input: ".$formalInputXML->getAttribute('Name');

		#######################################################################
		#
		# make OME::LookupTable & OME::LookupTable::Entry objects
		#
		#
		my $newLookupTable;
		my @lookupTables = $formalInputXML->getElementsByLocalName( "LookupTable" );
		# lookupTables may or may not exist. Either way is fine.
		if( scalar( @lookupTables ) == 1 ) {

			###################################################################
			#
			# make OME::LookupTable object
			#
			my $lookupTableXML = $lookupTables[0];
			my $descriptions = $lookupTableXML->getElementsByLocalName('Description');
			my $description = [$descriptions->[0]->childNodes()]->[0]->data()
				if $descriptions;

			my $data = {
				name        => $lookupTableXML->getAttribute( 'Name' ),
				description => $description
			};
			$newLookupTable = $factory->newObject( "OME::LookupTable", $data )
				or die "Could not create OME::LookupTable object\n";
			push(@commitOnSuccessfulImport, $newLookupTable);
			#
			###################################################################

			###################################################################
			#
			# make OME::LookupTable::Entry objects
			#
			my @entries = $lookupTableXML->getElementsByLocalName( "LookupTableEntry" );
			foreach my $entry (@entries) {
				my $data = {
					value           => $entry->getAttribute( 'Value' ),
					label           => $entry->getAttribute( 'Label' ),
					lookup_table_id => $newLookupTable->ID()
				};
				my $lookupEntry = $factory->newObject( "OME::LookupTable::Entry", $data )
					or die "Could not create OME::LookupTable::Entry object\n";
				push(@commitOnSuccessfulImport, $lookupEntry);
			}
			#
			###################################################################
		}
		#
		#
		#######################################################################

		#######################################################################
		#
		# make OME::FormalInput object
		#
		my $semanticType = $factory->findObject( "OME::SemanticType", name => $formalInputXML->getAttribute( 'SemanticTypeName' ) )
			or die "When processing Formal Input (name=".$formalInputXML->getAttribute( 'Name' )."), could not find Semantic type referenced by ".$formalInputXML->getAttribute( 'SemanticTypeName' )."\n";

		my ($optional, $list, $count);
		$count = $formalInputXML->getAttribute( 'Count' );
		if( $count ) {
			$optional = ( $count eq '*' || $count eq '?' ? 't' : 'f' );
			$list     = ( $count eq '*' || $count eq '+' ? 't' : 'f' );
		} 
		# FIXME: this is the wrong place to set default values. the default values for these columns are specified
		# in OME::Module::FormalInput, but something is broken. if I don't specify them here, then those columns
		# come up blank.
		else {
			$optional = 'f';
			$list = 't';
		}

		my $descriptions = $formalInputXML->getElementsByLocalName('Description');
		my $description = [$descriptions->[0]->childNodes()]->[0]->data()
			if $descriptions;

		my $data = {
			name               => $formalInputXML->getAttribute( 'Name' ),
			description        => $description,
			module_id         => $newProgram,
			semantic_type_id  => $semanticType->id(),
			lookup_table_id    => $newLookupTable,
			optional           => $optional,
			list               => $list,
			user_defined       => $formalInputXML->getAttribute( 'UserDefined' )
		};
		my $newFormalInput = $factory->newObject( "OME::Module::FormalInput", $data )
			or die ref ($self) . " could not create OME::Module::FormalInput object (name=".$formalInputXML->getAttribute( 'Name' ).")\n";

		push(@commitOnSuccessfulImport, $newFormalInput);
		$formalInputs{ $newFormalInput->name() } = $newFormalInput;
		#
		#
		#######################################################################

	}
	#
	#
	###########################################################################
	

	###########################################################################
	#
	# process formalOutputs
	#
	# this hash is keyed by FormalOutput.Name, valued by DBObject FormalOutput
	my %formalOutputs;

	logdbg "debug", ref ($self) . "->processDOM processing formal outputs";
	foreach my $formalOutputXML ( $moduleXML->getElementsByLocalName( "FormalOutput" ) ) {
		logdbg "debug", "\tformal output: ".$formalOutputXML->getAttribute('Name');

		###################################################################
		#
		# make OME::FormalOutput object
		#
        my $semanticTypeName = $formalOutputXML->getAttribute('SemanticTypeName');
        my $semanticType = $factory->findObject( "OME::SemanticType", name => $semanticTypeName );
        # Null semantic types are now allowed for formal outputs
        if (defined $semanticTypeName) {
            die "When processing Formal Output (name=".$formalOutputXML->getAttribute( 'Name' )."), could not find Semantic type referenced by ".$semanticTypeName."\n"
				unless defined $semanticType;
        }
		my ($optional, $list, $count);
		$count = $formalOutputXML->getAttribute( 'Count' );
		if( $count ) {
			$optional = ( $count eq '*' || $count eq '?' ? 't' : 'f' );
			$list     = ( $count eq '*' || $count eq '+' ? 't' : 'f' );
		} else {
                    $optional = 'f';
                    $list = 't';
                }

		my $descriptions = $formalOutputXML->getElementsByLocalName('Description');
		my $description = [$descriptions->[0]->childNodes()]->[0]->data()
			if $descriptions;
		my $data = {
			name               => $formalOutputXML->getAttribute( 'Name' ),
			description        => $description,
			module_id         => $newProgram,
			semantic_type_id  => $semanticType,
			feature_tag        => $formalOutputXML->getAttribute( 'IBelongTo' ),
			optional           => $optional,
			list               => $list
		};
		my $newFormalOutput = $factory->newObject( "OME::Module::FormalOutput", $data )
			or die "Could not create OME::Module::FormalOutput object\n";

		push(@commitOnSuccessfulImport, $newFormalOutput);
		$formalOutputs{ $newFormalOutput->name() } = $newFormalOutput;
		#
		###################################################################

	}
	#
	#
	###########################################################################
	
	
	# validate & process executionInstructions
	logdbg "debug", ref ($self) . "->processDOM processing ExecutionInstructions";
	my @executionInstructionList = 
		$moduleXML->getElementsByLocalName( "ExecutionInstructions" );
	if(scalar(@executionInstructionList) == 1) {
		my $executionInstructionsXML = $executionInstructionList[0];
		eval( "use $module_type" );
		die "module type $module_type cannot be loaded.\n$@" if $@;
		my $modifiedEIs = $module_type->
			validateAndProcessExecutionInstructions( $newProgram, $executionInstructionsXML );
        $newProgram->execution_instructions( $modifiedEIs->toString() )
        	if $modifiedEIs;
	}



	###########################################################################
	# commit this module. It's been successfully imported
	#
	logdbg "debug", ref ($self) . "->processDOM: imported module '".$newProgram->name."' sucessfully. Committing to DB...";
	while( my $DBObjectInstance = pop (@commitOnSuccessfulImport) ){
		$DBObjectInstance->storeObject();
	}                             # commits all DBObjects
    $session->commitTransaction();
	#
	###########################################################################
	
	push(@newPrograms, $newProgram)

} # END foreach my $moduleXML( @modules )
	
	return \@newPrograms;
	
} # END sub processDOM
#
#
###############################################################################


=pod

=head1 AUTHOR

Josiah Johnston (siah@nih.gov)

=head1 SEE ALSO

OME/src/xml/AnalysisModule.xsd - XML specification documents should conform to.
OME/src/xml/CLIExecutionInstructions.xsd - XML specification documents should conform to.

=cut


1;
