# OME/Tasks/ProjectManager.pm

#-------------------------------------------------------------------------------
#
# Copyright (C) 2003 Open Microscopy Environment
#       Massachusetts Institute of Technology,
#       National Institutes of Health,
#       University of Dundee
#
#
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#-------------------------------------------------------------------------------




#-------------------------------------------------------------------------------
#
# Written by:    J-M Burel <j.burel@dundee.ac.uk>
#
#-------------------------------------------------------------------------------


package OME::Tasks::ProjectManager;

=head1 NAME

OME::Tasks::ProjectManager - manage user's projects

=head1 SYNOPSIS

	use OME::Tasks::ProjectManager;
	my $projectManager = new OME::Tasks::ProjectManager($session);

=head1 DESCRIPTION

The OME::Tasks::ProjectManager provides a list of methods to manage user's projects

=head1 OBTAINING A PROJECTMANAGER

To retrieve an OME::Tasks::ProjectManager to use for managing the project, the user must log in to OME.  This is done via the L<OME::SessionManager|OME::SessionManager> class.  Logging in via OME::SessionManager yields an L<OME::Session|OME::Session> object.

	my $manager = OME::SessionManager->new();
	my $session = $manager->createSession($username,$password);
	my $projectManager = new OME::Tasks::ProjectManager;

=head1 METHODS (ALPHABETICAL ORDER)

The following methods are available to a "ProjectManager."

=head2 add ($datasetID,$projectID)

	$projectManager->add($dataset->id(), $project->id());

	$projectManager->add($dataset->id());

Adds the "Dataset" by reference of its ID to the "Project" by reference of its ID.

Note: The project ID is optional and if it is not defined the dataset is added to the current, active project ($session->Project());

=head2 addDatasets ($array_ref,$projectID)

	$projectManager->addDatasets([1,2,3], $project->id());

	$projectManager->addDatasets([1,2,3]);

Adds the "Datasets" provided by the array reference by reference of their ID to the "Project" by reference of its ID.

Note: The project ID is optional and if it is not defined the dataset is added to the current, active project ($session->Project());

=head2 change ($description,$name,$projectID)

	$projectManager->change(
		"My new description",
		"My new name",
		$project->id(),
	);

	$projectManager->change(
		"My new description",
		"My new name",
	);

Changes the "Project's" name and/or description by reference of its ID (or the current, active project if unspecified).

=head2 create ($hash_ref)

	$projectManager->create( {
		name => 'My New Project',
		description => 'A great new idea!',
		owner_id => $session->User()->ID(),
	});

Create a new project and update the OME session.

Note: This method is purely a macro for OME::Factory->newObject().

=head2 delete ($id)

	$projectManager->delete($project->id());

Delete a project and update OME session if the project is the current project.

Note: If the user doesn't have another project, the current active project and dataset are set to undefined. Otherwise the first arbitrary project and/or dataset are set to active.

=head2 getAllProjectCount ()

	my $p_count = $projectManager->getAllProjectCount();

	Gets a count of all the projects in the database.

=head2 getAllProjects ()

	my @projects = $projectManager->getAllProjects();

	Get all the projects in the database.

=head2 getAllProjectsLimit ($limit, $offset)

	my $p_count = $projectManager->getAllProjectCount();
	my @projects1 = $projectManager->getAllProjectsLimit($p_count / 2);
	my @projects2 = $projectManager->getAllProjectsLimit(
		$p_count / 2, $p_count - 100
	);

Retrieves a list of projects from the database using the LIMIT and OFFSET parameters to the query.

Note: The fields are ordered by their ID.

=head2 getDatasetCount ($project/$project_id)

	my $d_count1 = $projectManager->getDatasetCount($project1);
	my $d_count2 = $projectManager->getDatasetCount($project2->id());

	Get the dataset count of a given project.

Note: This method allows you to pass either a project object or a project ID.

=head2 getUserProjectCount ($experimenter)

	my $user_p_count = $projectManager->getUserProjectCount();
	my $other_p_count = $projectManager->getUserProjects(
		$other_experimenter
	);

	Get a count of all the projects owned by a given user.

Note: By default this method uses the Session's experimenter as a filter.

=head2 getUserProjects ($experimenter)

	my @user_projects = $projectManager->getUserProjects();
	my @other_user_projects = $projectManager->getUserProjects(
		$other_experimenter
	);

	Get all the projects owned by a given user.

Note: By default this method uses the Session's experimenter as a filter.

=head2 getUserProjectsLimit ([$experimenter], $limit, $offset)

	my $p_count = $projectManager->getAllProjectCount();
	my @projects1 = $projectManager->getUserProjectsLimit($p_count / 2);
	my @projects2 = $projectManager->getUserProjectsLimit(
		$p_count / 2, $p_count - 100
	);
	my @projects3 = $projectManager->getUserProjectsLimit( 
		$experimenter, $p_count / 2, $p_count - 100
	);

Retrieves a list of projects from the database using the LIMIT and OFFSET parameters to the query for a given user.

Note: By default this method uses the Session's experimenter as a filter. In addition, the fields are ordered by ID.

=head2 nameExists ($name)

	if ($projectManager->nameExists("A really good project name") {
		...

	} else {
		...
	}


Check if a given project name already exists in the database.

Returns successful (1) or unsuccessful (0) in matching.

=head2 listMatching ($userID,$array_ref)

	my $projects = $projectManager->listMatching();

	foreach (@$projects) {
		...
	}

Returns a list of all the project objects in the database.

	my $user_projects = $projectManager->listMatching(
		$session->User()->id(),
	);

	foreach (@$user_projects) {
		...
	}

Returns a list of all the project objects in the database owned by the specified user ID.

	my $related_projects = $projectManager->listMatching(
		$session->User()->id(),
		[ $ome_group_a->id(), $ome_group_b->id(), $ome_group_c->id() ],
	);

	foreach (@$related_projects) {
		...
	}

Returns a list of all the project objects in the database owned by the specified user ID or owned by one of the groups specified.

=head2 load ($projectID)

	my $project_a = $projectManager->load( 1 );
	my $project_b = $projectManager->load( 2 );

Returns a project object.

=head2 removeDatasets ($hash_ref)

	$projectManager->removeDatasets({
		$project1->id() => [1, 5, 10, 9],
		$project2->id() => [$dataset1, $dataset2],
	});

Project removal method which accepts a hash reference keyed by project
ID and containing an array reference of dataset objects or dataset ID's.
Results a session->commitTransaction.

Returns 1 on success and undef on failure, this is an *all or nothing method* (either every remove is successful or the entire task fails)

=head2 switch ($id,$bool)

	$projectManager->switch(1, 1);

Switches the current active project. If the boolean switch is active other checking will be done on the session in order to preserve dataset integrity.

=cut

use strict;
use OME::SetDB;
use OME::DBObject;
OME::DBObject->Caching(0);

use OME;
our $VERSION = $OME::VERSION;
use Carp;

sub new{
	my $class=shift;
	my $self={};
	
	return bless($self,$class);
}

###############################
# Parameters:
# 	datasetID = dataset_id to add
#	projectID = if not defined, add to current project

sub add{
	my $self=shift;
	my $session=$self->Session();
	my ($datasetID,$projectID)=@_;
	my $project;

	if (defined $projectID){
	    $project=$session->Factory()->loadObject("OME::Project",$projectID);
	}else{
	    $project=$session->project();
	}

	my $dataset=$self->addToProject($datasetID,$project->id());
	$dataset->storeObject();
	$session->commitTransaction();

	return $dataset;
}

###############################
# Parameters:
# 	datasetID = dataset_id to add
#	projectID = if not defined, add to current project

sub addDatasets {
	my $self=shift;
	my $session=$self->Session();
	my ($datasetIDs,$projectID)=@_;
	my $project;

	unless (defined $projectID){
	    $project=$session->project();
	}

	foreach (@$datasetIDs) {
		my $dataset=$self->addToProject($_,$projectID);
		$dataset->storeObject();
		$session->commitTransaction();
	}
}


#################
# Parameters: (void)
# 	
sub getAllProjects {
	my $self = shift;
	my $factory = $self->Session()->Factory();

	return $factory->findObjects("OME::Project");
}

#################
# Parameters: ($limit, $offset)
# 	
sub getAllProjectsLimit {
	my ($self, $limit, $offset) = @_;
	my $factory = $self->Session()->Factory();

	# Sanity
	return unless $limit;
	
	# Factory criterea specification
	my $criteria = {__limit => $limit , __order => ['id']};

	if ($offset) { $criteria->{'__offset'} = $offset; }

	return $factory->findObjects("OME::Project", $criteria);
}

#################
# Parameters: (void)
# 	
sub getAllProjectCount {
	my $self = shift;
	my $factory = $self->Session()->Factory();

	return $factory->countObjects("OME::Project");
}

#################
# Parameters: (project/project_id)
# 	
sub getDatasetCount {
	my ($self, $project_id) = @_;
	my $factory = $self->Session()->Factory();

	if (ref($project_id) eq 'OME::Project') {
		$project_id = $project_id->id()
	}

	return $factory->countObjects("OME::Project::DatasetMap", project_id => $project_id);
}

#################
# Parameters: (experimenter object)
#
sub getUserProjects {
	my ($self, $experimenter) = @_;
	my $factory = $self->Session()->Factory();

	$experimenter = $self->Session()->User() unless defined $experimenter;

	return $factory->findObjects("OME::Project", owner_id => $experimenter->id());
}

#################
# Parameters: ([experimenter object], limit, offset)
#
sub getUserProjectsLimit {
	my ($self, @fields) = @_;
	my $factory = $self->Session()->Factory();

	my ($experimenter, $limit, $offset);

	return if scalar(@fields) > 3;

	foreach (shift @fields) {
		if (ref($_) eq 'OME::SemanticType::__Experimenter') { 
			$experimenter = $_;
		} elsif (not defined $limit) {
			$limit = $_;
		} else {
			$offset = $_;
		}
	}

	$experimenter = $self->Session->User() unless $experimenter;

	return unless $limit;

	my $criteria = {
		owner_id => $experimenter->id(),
		__limit => $limit,
		__order => ['id'],
	};

	if ($offset) { $criteria->{'__offset'} = $offset; }

	return $factory->findObjects("OME::Project", $criteria);
}


#################
# Parameters: (experimenter object)
#
sub getUserProjectCount {
	my ($self, $experimenter) = @_;
	my $factory = $self->Session()->Factory();

	$experimenter = $self->Session()->User() unless defined $experimenter;

	return $factory->countObjects("OME::Project", owner_id => $experimenter->id());
}

#################
# Parameters
#	datasetId
#	projectID
sub addToProject{
    my $self=shift;
    my ($datasetID,$projectID)=@_;
    my $session=$self->Session();
    my $factory=$session->Factory();

    # maybeNewObject will work regardless of whether the values
    # are objects or ID's

    $factory->maybeNewObject("OME::Project::DatasetMap",
                             {
                              project_id => $projectID,
                              dataset_id => $datasetID,
                             });
    $session->commitTransaction();

    return ref($datasetID)?
      $datasetID:
      $factory->loadObject("OME::Dataset",$datasetID);
}

###############################
# Paramaters:
#	description = project's description 
#	name		= project's name
#	projectID	= projectId (optional)

sub change{
 	my $self=shift;
	my $session=$self->Session();
	my ($description,$name,$projectID)=@_;
	my $project;
	if (defined $projectID){
		$project=$session->Factory()->loadObject("OME::Project",$projectID);
	}else{
		$project=$session->project();
	}
	$project->name($name) if defined $name;
	$project->description($description) if defined $description;
	$project->storeObject();
	$session->commitTransaction();
	return 1;


}

#####################
# Parameters:
#	ref = project's informations

sub create{
	my $self=shift;
	my $session=$self->Session();
	my ($ref)=@_;
	$ref->{owner_id} = $session->experimenter_id()
		unless (exists $ref->{owner} and $ref->{owner})
		or (exists $ref->{owner_id} and $ref->{owner_id});
	$ref->{group_id} = $session->experimenter()->Group()->id()
		unless (exists $ref->{group} and $ref->{group})
		or (exists $ref->{group_id} and $ref->{group_id});
	my $project = $session->Factory()->newObject("OME::Project", $ref);

	$project->storeObject();
	$session->project($project);
	$session->storeObject();
	$session->commitTransaction();

	return 1;

}

######################
# Parameters:
#	id = project_id to delete

sub delete{
	my $self=shift;
	my $session=$self->Session();
	my ($id)=@_;
	my $result=undef;
	my $rep=undef;
	my $currentProject=$session->project();
	my $deleteProject=$session->Factory()->loadObject("OME::Project",$id);
	my @projects=$session->Factory()->findObjects("OME::Project",'owner_id'=>$session->User()->id() );
	 
	#my @datasets=$deleteProject->datasets();
	my @datasets=();
     	my @dMaps=$session->Factory()->findObjects(
			"OME::Project::DatasetMap",
			'project_id'=>$deleteProject->id()
		);
     	foreach my $d (@dMaps){
      		push(@datasets,$d->dataset());
    	} 

	my $db=new OME::SetDB();

	if (scalar(@datasets)>0){
	  $result=deleteProjectDatasetMap($deleteProject,\@datasets,$db);
	  return $result unless (defined $result);
	}

	if ($deleteProject->id()==$currentProject->id()){
	  if (scalar(@projects)==1){
	     $session->dataset(undef) if scalar(@datasets)>0;
	     $session->project(undef);
	     $session->storeObject();
		 $session->commitTransaction();
	  }else{
		$self->reorganizeSession($deleteProject,\@projects);
 	  }
	}

	$rep=deleteProject($deleteProject,$db);
	$db->Off();
	return $rep;
	 
}

################
# Parameters:
# name = project's name
# Return: 1 or 0

sub nameExists {
	my $self=shift;
	my $session=$self->Session();
	my ($name)=@_;
	my @list=$session->Factory()->findObjects("OME::Project",'name'=>$name);
	return scalar(@list) > 0 ? 1 : 0;
}



###############
# Parameters: 
#	userID (Optional)
#	$ref=ref array  list group_id (optional)
# Return: ref array of project objects owned by a given user.

sub listMatching{
	my $self=shift;
	my $session=$self->Session();
	my ($userID,$ref)=@_;
	my @projects=();
	if (defined $userID){
		@projects=$session->Factory()->findObjects("OME::Project",'owner_id'=>$session->User()->id());

	}else{
		if (defined $ref){
			foreach (@$ref){
			   push(@projects,$session->Factory()->findObjects("OME::Project",'group_id'=>$_));
			}
		
		}else{
			 @projects=$session->Factory()->findObjects("OME::Project");
		}
	}
	return \@projects;
}


############
# Parameters:
#	projectID =project_id to load
# Return: project object

sub load{
	my $self=shift;
	my $session=$self->Session();
	my ($projectID)=@_;
	my $project=$session->Factory()->loadObject("OME::Project",$projectID);
	return $project;
}

sub removeDatasets {
	my ($self, $to_remove) = @_;
	my $session = $self->Session();
	my $factory = $session->Factory();

	# Unless you give me something, go away
	return undef unless $to_remove;

	foreach my $project_id (keys %$to_remove) {
		my $dataset_ids = $to_remove->{ $project_id };
		foreach my $dataset_id ( @$dataset_ids ) {
			my $map = $factory->findObject("OME::Project::DatasetMap",
								{
								'dataset' => $dataset_id,
								'project' => $project_id,
								})
				or next;
			# We have the map. Delete it.
			$map->deleteObject();
		}
	}

   	$session->commitTransaction();
	return 1;
}

		

###############
# Parameters:
#	id = project_id
#	bool (optional) = check associated dataset

sub switch{
	my $self=shift;
	my $session=$self->Session();
	my ($id,$bool)=@_;
	my $project=$session->Factory()->loadObject("OME::Project",$id);
	$session->project($project);
	if (not defined $bool){
	    my @datasets=();
     	    my @dMaps=$session->Factory()->findObjects("OME::Project::DatasetMap",'project_id'=>$project->id() );
     	    foreach my $d (@dMaps){
      		 push(@datasets,$d->dataset());
    	    }



	  #my @datasets=$project->datasets();
	  if (scalar(@datasets)==0){
	   $session->dataset(undef); 		
	  }else{
	   $session->dataset($datasets[0]);
	  }
	}

	$session->storeObject();
	$session->commitTransaction();

	return 1;
}


###########################
# PRIVATE METHODS		  #
###########################

# METHODS DON'T USE delete function of Class::DBI

sub deleteProject{
	my ($deleteProject,$db)=@_;
	my $tableProject="projects";	
 	my ($condition,$result);
 	$condition="project_id=".$deleteProject->id();
 	$result=do_request($tableProject,$condition,$db);
 	return (defined $result)?1:undef;
}


sub deleteProjectDatasetMap{
  	my ($deleteProject,$ref,$db)=@_;
  	my $tableProjectMap="project_dataset_map";
  	my $result;
  	foreach (@$ref){
   	 my ($condition);
    	 $condition="project_id=".$deleteProject->id()." AND dataset_id=".$_->id();
    	 $result=do_request($tableProjectMap,$condition,$db);
    	 return undef if (!defined $result);
	
      }
  	return (defined $result)?1:undef;
}



sub reorganizeSession{
	my ($self,$deleteProject,$ref)=@_;
	my $session = $self->Session(); 
	my @new=();
	foreach (@$ref){
        push(@new,$_) unless $_->id()==$deleteProject->id();
	}
	my $newproject=$new[0];
	my @newdataset=();
     	my @dMaps=$session->Factory()->findObjects("OME::Project::DatasetMap",'project_id'=>$newproject->id() );
     	foreach my $d (@dMaps){
      	push(@newdataset,$d->dataset());
    	}


	#my @newdataset=$newproject->datasets();
	$session->project($newproject);
	if (scalar(@newdataset)==0){
		$session->dataset(undef);	
	}else{
		$session->dataset($newdataset[0]);
	}
	$session->storeObject(); 
	$session->commitTransaction();
	return 1;
}

# Macro to return the Session instance
sub Session { return OME::Session->instance() }

sub do_request{
 	my ($table,$condition,$db)=@_;
 	my $result;
 	if (defined $db){
      	$result=$db->DeleteRecord($table,$condition);
	 }
 	return $result;
}

=head1 AUTHOR

JMarie Burel (jburel@dundee.ac.uk)

=head1 SEE ALSO

L<OME::Project|OME::Project>,
L<OME::DBObject|OME::DBObject>,
L<OME::Factory|OME::Factory>,
L<OME::SetDB|OME::SetDB>,

=cut

1;

